"""
Some useful little functions
"""


def list_of_int(array_1d):
    """
    transforms all elements in a given 1d list/1d array into integers and returns a list.
    """
    return [int(x) for x in array_1d]
